package org.medical.services;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;

public class JavaRuleEngine {

    ArrayList<Rule> rules;
    int numberOfConditions = 8;
    int numberOfActions = 3;
    DateFormat df = new SimpleDateFormat("yyyy-MM-dd");

    public void setRules(String fileName) throws Exception {
        rules = new ArrayList<Rule>();
        String[] columns = null;
        String line = "";
        int l = 0;
        try {
            BufferedReader reader = new BufferedReader(new FileReader(fileName));
            boolean header = true;
            while ((line = reader.readLine()) != null) {
                l++;
                String[] elements = line.split(",");
                if (elements.length != (numberOfConditions + numberOfActions)) {
                    System.out.println("Invalid CSV file " + fileName + " line=" + l);
                    System.exit(-1);
                }
                if (header) {
                    header = false;
                    columns = elements;
                    continue;
                }
                Rule rule = new Rule(numberOfConditions, numberOfActions);
                rules.add(rule);
                for (int i = 0; i < numberOfConditions; i++) {
                    if (i < 6) { // strings
                        rule.setCondition(i, new ConditionString(columns[i], elements[i]));
                        continue;
                    }
                    Date date = null;
                    if (!elements[i].isEmpty())
                        date = df.parse(elements[i]);
                    if (i == 6) {
                        rule.setCondition(i, new ConditionDate(columns[i], "<=", date));
                    }
                    if (i == 7) {
                        rule.setCondition(i, new ConditionDate(columns[i], ">=", date));
                    }
                }
                for (int i = 0; i < numberOfActions; i++) {
                    int index = numberOfConditions + i;
                    rule.setAction(i, new Action(columns[index], elements[index]));
                }
            }
            reader.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public boolean execute(MedicalService service) {
        int n = 0;
        for (Rule rule : rules) {
            n++;
            if (rule.isSatisfied(service)) {
//                System.out.println("Rule " + n + " is satisfied");
                rule.apply(service);
                return true;
            }
        }
//        System.out.println("No rules satisfied for "+service);
        return false;
    }

    public void showRules() {
        int n = 0;
        for (Rule rule : rules) {
            n++;
            System.out.println(n + ": " + rule);
        }
    }

    public static void main(String[] args) {
        JavaRuleEngine engine = new JavaRuleEngine();

        try {
            engine.setRules("data\\MedicalServicesDecisionTable.csv");
        } catch (Exception e) {
            e.printStackTrace();
        }
        engine.showRules();
    }

    class Rule {
        Condition[] conditions;
        Action[] actions;

        public Rule(int numberOfConditions, int numberOfActions) {
            conditions = new Condition[numberOfConditions];
            actions = new Action[numberOfActions];
        }

        public void setCondition(int index, Condition cond) {
            conditions[index] = cond;
        }

        public void setAction(int index, Action action) {
            actions[index] = action;
        }

        public boolean isSatisfied(MedicalService service) {
            for (Condition c : conditions) {
                if ("Place Of Service".equals(c.getName()))
                    if (!c.compare(service.getPlaceOfService()))
                        return false;
                if ("Service Type".equals(c.getName()))
                    if (!c.compare(service.getType()))
                        return false;
                if ("Plan".equals(c.getName()))
                    if (!c.compare(service.getPlan()))
                        return false;
                if ("Group Size".equals(c.getName()))
                    if (!c.compare(service.getGroupSize()))
                        return false;
                if ("In Network".equals(c.getName()))
                    if (!c.compare(service.getInNetwork()))
                        return false;
                if ("Is Covered".equals(c.getName()))
                    if (!c.compare(service.getIsCovered()))
                        return false;
                if ("Service Type".equals(c.getName()))
                    if (!c.compare(service.getType()))
                        return false;
                if ("Effective Period Start".equals(c.getName()))
                    if (!c.compare(service.getDateOfService()))
                        return false;
                if ("Effective Period End".equals(c.getName()))
                    if (!c.compare(service.getDateOfService()))
                        return false;
            }
            return true;
        }
        
        public void apply(MedicalService service) {
            for (Action a : actions) {
                if ("Covered in Full".equals(a.getName())) {
                    service.setCoveredInFull(a.getValue());
                    continue;
                }
                if ("Copay".equals(a.getName())) {
                    service.setCopay(a.getValue());
                    continue;
                }
                if ("Coinsurance".equals(a.getName())) {
                    service.setCoInsurance(a.getValue());
                    continue;
                }
            }
        }

        @Override
        public String toString() {
            return "RuleJava [conditions=" + Arrays.toString(conditions) + ", actions=" + Arrays.toString(actions)
                    + "]";
        }

    }

    abstract public class Condition {
        String name;

        abstract public boolean compare(Object par);

        public String getName() {
            return name;
        }
    }

    class ConditionString extends Condition {

        String value;

        public ConditionString(String name, String value) {
            this.name = name;
            this.value = value;
        }

        public boolean compare(Object par) {
            if (value.isEmpty())
                return true;
            return value.equals(par);
        }

        @Override
        public String toString() {
            return name + "=" + value;
        }

    }

    class ConditionDate extends Condition {
        Date date;
        String oper;

        public ConditionDate(String name, String oper, Date date) {
            this.name = name;
            this.oper = oper;
            this.date = date;
        }

        public boolean compare(Object par) {
            if (date == null)
                return true;
            int comparison = date.compareTo((java.util.Date) par);
            if (oper.contentEquals("<=") && comparison <= 0)
                return true;
            if (oper.contentEquals(">=") && comparison >= 0)
                return true;
            return false;
        }

        @Override
        public String toString() {
            return name + oper + date;
        }
    }

    class Action {
        String name;
        String value;

        public Action(String name, String value) {
            this.name = name;
            this.value = value;
        }

        public void assign(String par) {
            value = par;
        }

        public String getName() {
            return name;
        }
        
        public String getValue() {
            return value;
        }

        @Override
        public String toString() {
            return name + "=" + value;
        }
    }
}